read("steinberg");
read("generators");
root = Strexpand("$HOME/tex/K2/src/new/");

/*
PRINT(x1,x2,x3,x4,x5,x6,x7)=;
PRINT1(x1,x2,x3,x4,x5,x6,x7)=;
*/
PRINT(v[..])=if (default("debug"), print(concat(v)));
PRINT1(v[..])=if (default("debug"), print1(concat(v)));

print_red(s)= PRINT("\e[0;91m", s, "\e[0m");

compute_bnf_if_needed(pol)=
{
  if (type(bnf)=="t_VEC", return);
  PRINT1("Computing bnf...");
  bnf = bnfinit( if (type(pol) == "t_INT", x^2 - pol, pol) );
  PRINT("done.");
}

GrowBound()=
{ my(rho, n, d, dtilde, sqrtD);

  n = poldegree(bnf.pol);
  rho = if (n == 2, Pi/sqrt(12),
            n == 3, Pi/sqrt(18), /* Hales */
            (n+2)/2 * 2^(-n/2)); /* Blichfeldt-Minkowski */
  sqrtD = sqrt(abs(bnf.disc));
  d = 2^n * gamma(n/2+1) / (Pi * n )^(n/2) * sqrtD;
  dtilde = (2/Pi)^(bnf.sign[2]) * sqrtD;
  vecmax([2^(2*n)*rho*d^2, 2^(2*n/3)*rho^(1/3)*(d*dtilde^2)^(2/3), rho*d^3]);
}

K2(p)=relations(p); generators();

\\ generator (polymod) of principal ideal. set piv and invpiv = 1/piv
genid(id)=
{ my(x);

  x = bnfisprincipal(bnf,id)[2];
  piv = nfbasistoalg(bnf,x);
  invpiv = 1 / piv;
  PRINT("    Prime: ", Prime);
  PRINT("    uniformizer: ", lift(piv));
}

/* compute a uniformizer pi_v for Prime above p in Z_{K,S}.
  * S = all prime ideals "less than Prime".
  * Return 0 if not principal, set piv and invpiv otherwise (in genid) */
Suniformizer(Prime)=
{ my(Nv, k, S, p, P, Plist, v);

  \\ already initialized ?
  if (type(piv) == "t_POLMOD" && piv.mod == bnf.pol, return);
  if (type(Prime) == "t_INT", Prime = idealprimedec(bnf,Prime)[1]);
 
  p = Prime.p;
  Nv= idealnorm(bnf,Prime);
  v = bnfisprincipal(bnf,Prime,0);
  if (!v, genid(Prime); return); \\ Prime is principal
  S = [Prime]~;
  v = Mat(v);
  forprime(l=2, p-1,
    Plist = idealprimedec(bnf,l);
    \\ invert all prime ideals above p but one
    for (i=1, #Plist - 1, 
      P = Plist[i];
      if (idealnorm(bnf,P) >= Nv, break);
     
      S = concat(S, [P]~);
      v = concat(v, bnfisprincipal(bnf, P, 0));
      k = matkerint(v);     \\ seek relation c_1 + sum a_ic_i = 0
      for (i=1, #k,
        if (abs(k[1,i]) == 1,
          P = idealfactorback(bnf, S, k[,i]);
          if (k[1,i] < 0, P = idealinv(bnf,P));
          genid(P); return
        )
      )
    )
  ); piv = 0;
}
