\\ $Id: steinberg.gp,v 1.1.1.1 2006/08/10 17:33:43 kb Exp $
install(mathnfspec, "G&&&&", hnfspec);
install(hnfadd, "GG&&&GG");

global(candmax = 10000);
\\ list of elements to avoid (as t_STR)
global(already = List());
\\ list of elements found so far and not treated (in update_*) yet
global(curlist = List());

\\ fundamental S-units
global(Wall);

\\ relation matrix, components of its hnf (see hnfadd/spec)
global(matwedge, globalperm, globalB);

global(DEBUG = 0);

Zissmooth(n)=
{ my(no = abs(n));
  no==1 ||  no == prod(j=1, lp, listp[j]^valuation(no,listp[j]))
}

issmooth(x)= Zissmooth(norm(x))

\\ check that x is only divisible by primes in S
issmooth(x)=
{ my(f);

  if (!Zissmooth(norm(x)), return(0));
  f = idealfactor(bnf,x)[,1];
  # setminus(Set(f), SSet) == 0
}

decompose(alpha)= lift(bnfissunit(bnf,bnfS,alpha))

/* check if `a' or one of 5 related expressions belongs to already. Remember
 * it if not. */
if_new_remember(a)=
{ my(e);
  
  ind = setsearch(already, a, 1);
  if (ind == 0
      || setsearch(already, 1 - a)
      || setsearch(already, e = 1/a)
      || setsearch(already, 1 - e)
      || setsearch(already, e = 1/(1-a))
      || setsearch(already, 1 - e), return);
  listinsert(already, Str(a), ind);
  listput(curlist, a); triples++;
}

append_candnum(a,bmax)=
{ my(alpha);
 
  for (b = if (!a,1,0), bmax,
    alpha = nfbasistoalg(bnf, [a,b]~);
    if (Zissmooth(norm(alpha)), lauf1++;
      candnum[lauf1] = alpha;
      candcont[lauf1] = gcd(a,b);
      if (lauf1 == maxlauf1, return(1))
    )
  )
}

append_candnum(a,bmax)=
{ my(alpha,v);
 
  forvec (b = if(!a, bmax0, bmax),
    v = concat([a],b)~;
    alpha = nfbasistoalg(bnf, v);
    if (issmooth(alpha), lauf1++;
      candnum[lauf1] = alpha;
      candcont[lauf1] = content(v);
      if (lauf1 == maxlauf1, return(1))
    )
  )
}

init_matwedge(init) =
{ my(w,t, vecwedge, tu);
 
  if (!init, matwedge=[;]; return (0));
 
  w = bnf.tu[1]; tu = bnf.tu[2];
  t = #Wall;
  vecwedge = concat(concat(
      vector(t,  i, symbol(Wall[i],-Wall[i])), 
      vector(t,  i, 2 * symbol(Wall[i], Wall[i]))
    ), 
    vector(t-1,i, w * symbol(tu, Wall[i+1]))
  );
  matwedge = Mat(vecwedge); 0
}

\\ sort ideals by increasing norm
sortS(S)=
{ my(N);

  N = vector(#S,i,idealnorm(bnf,S[i]));
  vecextract(S, vecsort(N,,1));
}

RefinedGrowBound(atleast) =
{ my(lim,Prime,midpoint,S1,S2, T0);

  T0 = gettime();
  lim = floor(GrowBound());
  PRINT("Refining S starting from Groenewegen's bound: ", lim);
  triples = 0;
  initS(lim, atleast);
  PRINT1("initially, #S = "); print_red(#S);
  S = sortS(S);
 
  until (#S <= atleast,
    Prime = S[#S];
    S = vecextract(S, "..-2"); \\ last but one
    do_init = 1; initW(S);
    clear_lists();
    if (!is_U1_Ker_beta(Prime), break);
    PRINT1("\n#S = "); print_red(#S)
  );
 
  print_red("Done!");
  PRINT("time = ", gettime() - T0, "ms");
  return (idealnorm(bnf,Prime))
}

get_dim() = #Wall * (#Wall+1) / 2

clear_lists() = 
{
  already = List();
  curlist = List();
}

\\ if do_init (boolean) is unset don't execute
init_relations1(pol, limW, initmatwedge)=
{ my(lim, atleast);

  if (!do_init, return);
 
  compute_bnf_if_needed(pol);
  atleast = 10;
  if (limW < 0, atleast = -limW; limW = 2);
  if (!limW, limW = RefinedGrowBound(atleast));
  clear_lists();
  initS(limW, atleast);
  PRINT1("last norm = ", idealnorm(bnf,S[#S]), ", limW = ", limW, ", #S = "); print_red(#S);
  initW(S);
  init_matwedge(initmatwedge);
  triples = last_index = do_init = 0;
}

init_relations2()=
{ my(amax,bmax,l);

  amax = abs(bnf.disc) * lp;
\\bmax = 30;
  bmax = vector(poldegree(bnf.pol)-1, i, [-3,3]);
  bmax0= concat([[1,3]], vector(poldegree(bnf.pol)-2, i, [-3,3]));
  lauf1=0; maxlauf1 = max(4*get_dim(), 100);
  candnum  = vector(maxlauf1);
  candcont = vector(maxlauf1);
  PRINT("looking for ", maxlauf1, " candidates");
  if (!append_candnum(0,bmax),
    for (a = 1, amax,
      if (append_candnum( a,bmax), break);
      if (append_candnum(-a,bmax), break);
    )
  );
  candnum = candnum[1..lauf1];
  ind = vecsort(norm(candnum),,1);
  l = #candnum;
  candnum  = vector(l, i,  candnum[ind[i]]);
  candcont = vector(l, i, candcont[ind[i]]); \\ sort candnum accord. to norm
  PRINT("lauf1 = ", lauf1);
}

\\ produces vector  listp  of rational primes p s.t there exists P | p,
\\ NP <= Nv and  vector  S  of such prime ideals P.
initS(Nv, atleast)=
{ my(P,lS);

  S = listp = [];
  forprime(p=2, Nv,
    lS = #S;
    P = idealprimedec(bnf,p);
    for (i=1, #P,
      if (idealnorm(bnf,P[i]) <= Nv, S = concat(S, [P[i]]))
    );
    if (#S == lS, continue);
    listp = concat(listp,[p]);
    if (atleast < 0 && #S > -atleast, break)
  );
  atleast = abs(atleast);
  lp = #listp;
  if (#S < atleast, initS(4*Nv, -atleast), S)
}

\\  compute an S-unit basis S = { \wp, norm(\wp) <= lim }
initW(S)=
{
  bnfS = bnfsunit(bnf, S);
  \\fund .units, torsion unit, non-units
  Wall = matbasistoalg(bnf, concat(bnf.futu, bnfS[1]));
  SSet = Set(S);
  0
}

relations(pol, limW, extra_rel = 40, jinit = 2)=
{ local(first);
  \\ extra_rel: compute HNF of matwedge (relation matrix) when dim + extra_rel
  \\ relators have been obtained (maximal rank = dim expected).

  do_init = (jinit==2);
  if (do_init, 
    if (pol, bnf = 'bnf); \\ reset bnf
    init_relations1(pol, limW, 1);
    init_relations2();
    searchmultiplicative();
    first = 1;
    if (triples > get_dim(), update_matwedge());
    enough_rel = get_dim() + extra_rel;
  ,
    PRINT("looking for extra relations...");
    first = 0;
    lp = #listp;
    enough_rel = triples + extra_rel;
  );
  miss = 1;

  for (j=jinit, lauf1,
    cn = candnum[j];
    for (k=1, j-1,
      if (gcd(candcont[j],candcont[k])==1,
        cd = candnum[k];
        cq = cn / cd;
        p = cd - cn;
        if (issmooth(p), if_new_remember(cq));
        p = cd + cn; cq = -cq;
        if (p && issmooth(p), if_new_remember(cq));
        if (triples >= enough_rel,
          update_matwedge();
          if (!miss, break(2));
          last_index += triples;
          enough_rel =  miss + extra_rel;
          if (last_index > candmax - enough_rel,
          \\ too many elements remembered. Delete them (STUPID)
            listkill(already); last_index = 0;
          )
        )
      )
    );
    PRINT1("(", j, ", ", triples, ") ");
    jinit = j;
  );
  if (miss, update_matwedge());
  if (miss,
    print_red("\nEnlarging Factorbase...");
    relations(pol, -(#S + 2), extra_rel)
  ,
    PRINT("determinant: ", matdet(matwedge));
  ); jinit
}

check(a, ww, zet)=
{ my(z = a);
  for (j=1,ww,
    z *= zet;
    if (z != 1 && issmooth(1 - z), if_new_remember(z))
  )
}

checkdec(a, ww, zet)=
{ my(z = a, d);
  for (j=1,ww,
    z *= zet;
    if (z != 1 && (d = bnfissunit(bnf,bnfS,1-z)),
      triples++;
      listput(curlist, d)
    )
  )
}

searchmultiplicative()=
{ my(c, lg, zet, gen, z);

  ww = bnf.tu[1]; zet= bnf.tu[2];
  lg = #Wall;
  gen = vector(lg,j,Wall[lg+1-j]);
  c = 500;
  PRINT("multiplicative search");
  for(i=1, lg,
    check(gen[i], ww,zet);
    for (j=i+1, lg,
      check(gen[i] * gen[j], ww,zet);
      check(gen[i] / gen[j], ww,zet);
      if (triples >= c, return);
    )
  );
  PRINT("\ntriples = ", triples)
}

searchmultiplicative_for_U1(Prime)=
{ my(Nv);

  if (!do_init, return);
  PRINT("\nMultiplicative search for U1");
 
  if (type(Prime) == "t_INT", Prime = idealprimedec(bnf,Prime)[1]);
  Nv = idealnorm(bnf, Prime);
  init_relations1(bnf.pol, Nv-1, 0);
  is_U1_Ker_beta(Prime);
}

vconcat(a,b) = concat(a~, b~)~

\\ computes the full (HNF reduced) relation matrix from its parts
rebuild_rel(H) = 
{ my(top, bot, l,h);

  top = concat(vconcat(dep, H), globalB);
  h = matsize(globalB)[2];
  l = matsize(H)[1];
  bot = concat(matrix(h,l), matid(h));
  vconcat(top, bot);
}

is_U1_Ker_beta(Prime, extra_rel = 20)=
{ my(lg,ww,zet,gen,Nv,a,b);
  local(first = 1);
  local(dep,C,matz,Dmatz,enough_rel);
 
  piv = 'piv; \\ to make sure it is recomputed
  Suniformizer(Prime);
  if (!piv, PRINT("     not principal"); return(0));
 
  ww = bnf.tu[1]; zet = bnf.tu[2];
  lg = #Wall;
  gen = vector(lg,j,Wall[lg+1-j]); \\ rank increases much faster this way
  enough_rel = lg + extra_rel; \\ number of relations desired (rank = lg)
 
  matz = vectorv(lg); matz[#bnf.futu] = ww; \\ relation torsion part: zet^ww
  matz = Mat(matz);
  Nv = idealnorm(bnf,Prime);
  for(i=1, lg,
    a = piv * gen[i]; checkdec(a, ww, zet);
    b = piv / gen[i]; checkdec(b, ww, zet);
    for (j=i+1, lg,
      checkdec(a * gen[j], ww, zet);
      checkdec(b * gen[j], ww, zet);
      if (triples >= enough_rel && update_matz(Nv, extra_rel), return (1))
    )
  );
  if (!Dmatz,
    if (!triples,
      PRINT("no triples found!");
      U1 = 'U1; return
    );
    if (update_matz(Nv, extra_rel), return (1));
  );
  PRINT("U1 computed, but smaller than Ker(beta).");
  U1 = rebuild_rel(matz); 0
}

update_relations(x, new) = 
{
  if (#new <= 1, return(x));
  if (first, first = 0;
    x = concat(x,new);
    C = matrix(0,#x);
    x = hnfspec(x, globalperm, dep, globalB, C);
\\PRINT("hnfspec: "x)
  ,
    x = hnfadd(x, globalperm, dep, globalB, C, new, matrix(0,#new));
\\PRINT("hnfadd: "x)
  );
  miss = matsize(if (#dep, dep, globalB))[1]; /* missing relations */
  if (DEBUG && miss, PRINT ("missing relations = ", miss));
  x
}

update_matz(Nv, extra_rel)=
{ my(new, miss);

  new = Mat(curlist);
  curlist = List();
  new[#bnf.futu,] = lift(new[#bnf.futu,]); \\ lift torsion part
  matz = update_relations(matz, new);
  triples = 0;

  if (!miss,
    Dmatz = matdet(matz);
    if (Dmatz == Nv - 1,
      PRINT("U1 = Ker(Beta) ==> delta_v isomorphism ! Aborting...");
      U1_is_ker = 1; return(1)
    );
    PRINT ("[U:U1] = ", Dmatz)
  );
  enough_rel = miss + extra_rel;
  if (DEBUG,
    PRINT ("not enough. Going for ", enough_rel, " more");
  );
  0
}

update_matwedge()=
{ my(new, dim = get_dim());

  PRINT("\nchecking matwedge [", dim, "]");
  vecwedge = vector(triples,i, symbol(curlist[i], 1-curlist[i]));
  listkill(curlist);
  new = matrix(dim,triples,j,k, vecwedge[k][j]);
  matwedge = update_relations(matwedge, new);
  triples = 0;
}

symbol(x,y) =
{ my(u, v, w, L, lauf);

  L = #Wall;
  u = decompose(x);
  v = decompose(y); w = vectorv( get_dim() );
  for(i=1, L,
    w[lauf++] = u[i]*v[i] % 2;
    for(j=i+1, L,
      w[lauf++] = u[i]*v[j] - u[j]*v[i]
    )
  ); w
}

\\ x a factorization matrix of symbols. Return symbol in decomposed form.
Symbol(x) = sum(i=1, matsize(x)[1], x[i,2] * symbol(x[i,1][1], x[i,1][2]))
